/*
 * readinfo.cc 
 * Reads an NFO file into an array of parsed sprites.
 * 
 * Bastard child of NFORenum's renum.cpp, pseudo.cpp, and inlines.h
 *
 * Copyright 2004-2007 by Dale McCoy.
 * dalestan@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
INFO VERSION HISTORY:
Version 1: First version
Version 2: Add sprite numbers, // (Info version x) line
Version 3: Change real-sprite format
Version 4: Fix bug in real-sprite format
Version 5: Add literal strings
Version 6: Add binary includes
Version 7: Add backslash escapes
*/


#include<iostream>
#include<string>
#include<sstream>

using namespace std;

#include<boost/date_time/gregorian/gregorian_types.hpp>
using namespace boost::gregorian;

#include"nfosprite.h"
#include"allocarray.h"

extern uint nfogrfid;

istream&eat_white(istream&in){
	while(isspace(in.peek()))in.ignore();
	return in;
}

int ctoi(char ch){
	if(ch>='0'&&ch<='9')return ch-'0';
	if(ch>='A'&&ch<='F')return ch-'A'+10;
	if(ch>='a'&&ch<='f')return ch-'a'+10;
	return 0;
}

string UCase(string str){
	size_t len=str.length();
	for(size_t i=0;i<len;i++)
		str[i]=(char)toupper(str[i]);
	return str;
}

string itoa(uint x){
	string ret;
	while(x){
		ret="0123456789"[x%10]+ret;
		x/=10;
	}
	return ret;
}

uint ReadHex(istream&in,uint digits){
	uint ret;
	char ch;
	eat_white(in).get(ch);
	if((ret=ctoi(ch))==0&&ch!='0'){
		in.unget().clear(ios::badbit);
		return ret;
	}
	for(;--digits;){
		in.get(ch);
		if(ctoi(ch)==0&&ch!='0'){
			in.unget();
			return ret;
		}
		ret<<=4;
		ret|=ctoi(ch);
	}
	return ret;
}

const char*const VALID_PSEUDO="0123456789ABCDEFabcdef \t\v\r\n",
	*const WHITESPACE=" \t\v\r\n",
	*const COMMENT="#;/";

#define NPOS (string::npos)

#define flush_buffer()\
	if(buffer!=""){\
		sprites.push_back(Pseudo(sprites.size(),infover,buffer));\
		buffer="";\
	}else\
		(void(0))

void read_file(istream&in,AllocArray<Sprite>&sprites){

	string sprite,datapart,buffer;
	int infover;

	getline(in,buffer);		// read first line, a comment

	if (strncmp(buffer.c_str(), "// ", 3)) {
		throw 1;
	}
	getline(in,buffer);		// second line used to get info version
	if (strncmp(buffer.c_str(), "// (Info version ", 17)) {
		infover = 1;
		in.seekg(0);
	} else
		infover = atoi(buffer.c_str() + 17);

	int temp;
	string::size_type firstnotpseudo;
	buffer="";
	while(true){
		getline(in,sprite);
		istringstream spritestream(sprite);
		eat_white(spritestream);
		if(spritestream.peek()==EOF || // blank
			strchr(COMMENT,spritestream.peek())){ // comment
		}else{//sprite
			if(!eat_white(spritestream>>temp))
				spritestream.clear();
			if(spritestream.peek()=='*'){
				if(spritestream.ignore().peek()=='*'){
					flush_buffer();
					getline(eat_white(spritestream.ignore()),datapart);
					sprites.push_back(Include());
				}else{
					flush_buffer();
					eat_white(spritestream>>temp);
					getline(spritestream,buffer);
				}
			}else{
				getline(spritestream,datapart);
				firstnotpseudo=datapart.find_first_not_of(VALID_PSEUDO);
				if((!spritestream||firstnotpseudo==NPOS||
					(datapart[firstnotpseudo]=='"'&&infover>4)||
					(datapart[firstnotpseudo]=='\\'&&infover>6)||
					strchr(COMMENT,datapart[firstnotpseudo]))&&
					Pseudo::MayBeSprite(buffer)){
						//nothing
				}else{
					flush_buffer();
					sprites.push_back(Real());
				}
			}
		}
		if(in.peek()==EOF){
			flush_buffer();
			return;
		}
	}
}

Sprite::unparseable::unparseable(string reason,size_t sprite){
	this->reason="Error: "+reason+".\n\tWhile reading sprite:"+itoa((int)sprite)+'\n';
}

#define CHAR(x) (char(((ch>>((x)*6))&0x3F)|0x80))

string GetUtf8Encode(uint ch){
	if(ch<0x80)return string()+char(ch);
	if(ch<0x8000)return string()+char(((ch>>6 )&0x1F)|0xC0)+CHAR(0);
	/*if(ch<0x10000)*/return string()+char(((ch>>12)&0x0F)|0xE0)+CHAR(1)+CHAR(0);
	//if(ch<0x200000)return string()+char(((ch>>18)&0x07)|0xF0)+CHAR(2)+CHAR(1)+CHAR(0);
	//if(ch<0x4000000)return string()+char(((ch>>24)&0x03)|0xF8)+CHAR(3)+CHAR(2)+CHAR(1)+CHAR(0);
	//if(ch<0x80000000)return string()+char(((ch>>30)&0x01)|0xFC)+CHAR(4)+CHAR(3)+CHAR(2)+CHAR(1)+CHAR(0);
	//INTERNAL_ERROR(ch,ch);
}

#undef CHAR

Pseudo::Pseudo(size_t num,int infover,const string&sprite){
	istringstream in(sprite);
	ostringstream out;
	char ch;
	while(in){
		eat_white(in);
		switch(in.peek()){
		case EOF:continue;
		case'"':
			in.ignore();
			while(true){
				if(!in.get(ch))
					throw Sprite::unparseable("Unterminated literal string",num);
				if(ch=='"')
					break;
				if(ch=='\\'&&infover>6){
					switch(ch=(char)in.get()){
					case'n':
						ch='\r';//TTD uses Mac-style linefeeds (\r)
						break;
					case'"':
					case'\\':
						break;
					case'U':{
						uint x=ReadHex(in,4);
						if(!in)
							throw Sprite::unparseable("Could not parse quoted escape sequence",num);
						if(x>0x7FF)
							out.write(GetUtf8Encode(x).c_str(),3);
						else if(x>0x7F)
							out.write(GetUtf8Encode(x).c_str(),2);
						else out.put(x);
						continue;
						}
					default:
						in.unget();
						ch=(char)ReadHex(in,2);
						if(!in)
							throw Sprite::unparseable("Could not parse quoted escape sequence",num);
						break;
					}
				}
				out.put(ch);
			}
			break;
		case'/':case'#':case';'://comment
			in.ignore(INT_MAX,'\n');
			break;
		case'\\':
			if(infover>6){
				in.ignore();
				uint x;
				switch(in.get()){
				case'2':{
					bool sign=true;
					if(in.peek()=='u'){
						sign=false;
						in.ignore();
					}
					switch(in.get()){
					case'+':out.put(0x00);continue;
					case'-':out.put(0x01);continue;
					case'<':out.put(sign?0x02:0x04);continue;
					case'>':out.put(sign?0x03:0x05);continue;
					case'/':out.put(sign?0x06:0x08);continue;
					case'%':out.put(sign?0x07:0x09);continue;
					case'*':out.put(0x0A);continue;
					case'&':out.put(0x0B);continue;
					case'|':out.put(0x0C);continue;
					case'^':out.put(0x0D);continue;
					}
					break;
				}case'7':{
					switch(in.get()){
					case'1':out.put(0x00);continue;
					case'0':out.put(0x01);continue;
					case'=':out.put(0x02);continue;
					case'!':out.put(0x03);continue;
					case'<':out.put(0x04);continue;
					case'>':out.put(0x05);continue;
					case'G':
						if(in.get()=='G'){// \7GG Is or will be active
							out.put(0x09);
							continue;
						}
						in.unget();
						out.put(0x06);// \7G Is active
						continue;
					case'g':
						switch(in.get()){
						case'g':out.put(0x0A);continue;// \7gg is not and will not be active
						case'G':out.put(0x08);continue;// \7gG is not active but will be
						default:// \7g is not active
							in.unget();
							out.put(0x07);
							continue;
						}
					case'C':out.put(0x0B);continue;
					case'c':out.put(0x0C);continue;
					}
					break;
				}case'D':{
					bool sign=true;
					if(in.peek()=='u'){
						sign=false;
						in.ignore();
					}
					switch(in.get()){
					//<operation>
					case'=':out.put(0x00);continue;
					case'+':out.put(0x01);continue;
					case'-':out.put(0x02);continue;
					case'*':out.put(sign?0x04:0x03);continue;
					case'<':
						if(in.get()=='<'){
							out.put(sign?0x06:0x05);
							continue;
						}
						break;
					case'|':out.put(0x07);continue;
					case'&':out.put(0x08);continue;
					case'/':out.put(sign?0x0A:0x09);continue;
					case'%':out.put(sign?0x0C:0x0B);continue;
					//GRM <operation>
					case'R':out.put(0x00);continue;
					case'F':out.put(0x01);continue;
					case'C':out.put(0x02);continue;
					case'M':out.put(0x03);continue;
					case'n'://no fail
						switch(in.get()){
						case'F':out.put(0x04);continue; //\DnF
						case'C':out.put(0x05);continue; //\DnC
						}
						break;
					case'O':out.put(0x06);continue;
					}
				}case'b':
					if(in.peek()=='*'){// \b*
						in.ignore()>>x;
						if(!in||x>0xFFFF)break;//invalid
						if(x>0xFE){
							out.put('\xFF');
							out.put(x);
							out.put(x>>8);
						}else out.put((char)x);
						continue;
					}
					in>>x;
					if(x>=1920)x-=1920;//remap years
					if(!in||x>0xFF)break;//invalid
					out.put((char)x);
					continue;
				case'w':
					if(in.peek()=='x'){// \wx
						in.ignore();
						x=ReadHex(in,4);
					}else{
						in>>x;
						//delay fail check until after date parsing.
						if(in.peek()=='/'||in.peek()=='-'){//date
							in.ignore();
							unsigned short y,z;
							in>>y;
							if(/*!in||*/(in.peek()!='/'&&in.peek()!='-'))//in.peek will return eof if !in
								break;
							in.ignore();
							in>>z;
							if(x==0)x=2000;
							if(x>31&&x<100)x+=1900;
							if(z==0)z=2000;
							if(z>31&&z<100)z+=1900;
							date d;
							try{
								if(x>1919)d=date((unsigned short)x,y,z);
								else d=date(z,y,(unsigned short)x);
							}catch(std::out_of_range){
								break;
							}
							x=(d-date(1920,1,1)).days();
						}
					}
					if(!in)break;
					if(x>0xFFFF)break;//invalid
					out.put(x);
					out.put(x>>8);
					continue;
				case'd':
					if(in.peek()=='x'){// \dx
						in.ignore();
						x=ReadHex(in,8);
					}else in>>x;
					if(!in)break;
					out.put(x);
					out.put(x>>8);
					out.put(x>>16);
					out.put(x>>24);
					continue;
				}
				throw Sprite::unparseable("Could not parse unquoted escape sequence",num);
			}
		default:
			ch=(char)ReadHex(in,2);
			if(!in)
				throw Sprite::unparseable("Encountered invalid character looking for literal byte",num);
			out.put(ch);
		}
	}
	packed=out.str();
	if(!size())
		throw Sprite::unparseable("Found a zero-byte pseudo-sprite",num);
	if((*this)[0] == 8 && size() >= 8){
		nfogrfid = packed[2]<<24;
		nfogrfid |= packed[3]<<16;
		nfogrfid |= packed[4]<<8;
		nfogrfid |= packed[5];
	}
}

uint Pseudo::size()const{return (uint)packed.size();}

bool Pseudo::MayBeSprite(const string&sprite){
	istringstream in(sprite);
	char ch;
	while(in.get(ch)){
		if(ch=='"')return true;
		if(strchr(COMMENT,ch)){
			in.ignore(INT_MAX,'\n');
			continue;
		}
		if(isspace(ch)||string(VALID_PSEUDO).find(ch)==NPOS)continue;
		return true;
	}
	return false;
}
